/*
 * Copyright 2009 Red Hat, Inc.
 * Red Hat licenses this file to you under the Apache License, version
 * 2.0 (the "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 * implied.  See the License for the specific language governing
 * permissions and limitations under the License.
 */
package org.hornetq.jms.example;

import java.util.Enumeration;

import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.MessageConsumer;
import javax.jms.MessageProducer;
import javax.jms.Queue;
import javax.jms.QueueBrowser;
import javax.jms.Session;
import javax.jms.TextMessage;
import javax.jms.Topic;
import javax.jms.JMSException;

import org.hornetq.api.core.DiscoveryGroupConfiguration;
import org.hornetq.api.jms.HornetQJMSClient;
import org.hornetq.api.jms.JMSFactoryType;
import org.hornetq.common.example.HornetQExample;

/**
 * This example demonstrates a cluster of six nodes set up in a symmetric topology - i.e. each node
 * is connected to every other node in the cluster.
 * 
 * This is probably the most obvious clustering topology and the one most people will be familiar with from
 * using clustering in an app server, where every node has pretty much identical configuration to every other node.
 * 
 * By clustering nodes symmetrically, HornetQ can give the impression of clustered queues, topics and
 * durable subscriptions.
 * 
 * Please see the readme.html file for more information.
 *
 * @author <a href="tim.fox@jboss.com>Tim Fox for original SymmetricClusterExample</a>
 */
public class SymmetricClusterExample extends HornetQExample
{
   public static void main(final String[] args)
   {
      String[] serverArgs = new String[] {"-Xms50M",
                                          "-Xmx100M",
                                          "-XX:+UseParallelGC",
                                          "-XX:+AggressiveOpts",
                                          "-XX:+UseFastAccessorMethods" };

      new SymmetricClusterExample().run(args);
   }

   @Override
   public boolean runExample() throws Exception
   {
      Connection connection0 = null;
      Connection connection1 = null;
      Connection connection2 = null;
      Connection connection3 = null;
      Connection connection4 = null;
      Connection connection5 = null;
      Session session0 = null;
      Session session1 = null;
      Session session2 = null;
      Session session3 = null;
      Session session4 = null;
      Session session5 = null;
      MessageConsumer consumer0 = null;
      MessageConsumer consumer1 = null;
      MessageConsumer consumer2 = null;
      MessageConsumer consumer3 = null;
      MessageConsumer consumer4 = null;
      MessageConsumer consumer5 = null;
      Queue queue = null;

      try
      {
         // Step 1 - We instantiate a connection factory directly, specifying the UDP address and port for discovering
         // the list of servers in the cluster.
         // We could use JNDI to look-up a connection factory, but we'd need to know the JNDI server host and port for
         // the
         // specific server to do that, and that server might not be available at the time. By creating the
         // connection factory directly we avoid having to worry about a JNDI look-up.
         // In an app server environment you could use HA-JNDI to lookup from the clustered JNDI servers without
         // having to know about a specific one.
         DiscoveryGroupConfiguration groupConfiguration = new DiscoveryGroupConfiguration("231.7.7.7", 9876);
         
         ConnectionFactory cf = (ConnectionFactory)HornetQJMSClient.createConnectionFactoryWithHA(groupConfiguration, JMSFactoryType.CF);

         // We give a little while for each server to broadcast its whereabouts to the client
         Thread.sleep(2000);

         // Step 2. Directly instantiate JMS Queue object
         queue = HornetQJMSClient.createQueue("exampleQueue");

         // Step 3. We create six connections, they should be to different nodes of the cluster in a round-robin fashion
         // and start them
         connection0 = cf.createConnection();
         connection1 = cf.createConnection();
         connection2 = cf.createConnection();
         connection3 = cf.createConnection();
         connection4 = cf.createConnection();
         connection5 = cf.createConnection();

         connection0.start();
         connection1.start();
         connection2.start();
         connection3.start();
         connection4.start();
         connection5.start();

         // Step 4. We create a session on each connection

         session0 = connection0.createSession(false, Session.AUTO_ACKNOWLEDGE);
         session1 = connection1.createSession(false, Session.AUTO_ACKNOWLEDGE);
         session2 = connection2.createSession(false, Session.AUTO_ACKNOWLEDGE);
         session3 = connection3.createSession(false, Session.AUTO_ACKNOWLEDGE);
         session4 = connection4.createSession(false, Session.AUTO_ACKNOWLEDGE);
         session5 = connection5.createSession(false, Session.AUTO_ACKNOWLEDGE);

         // Step 6 has been relocated until after messages sent so the consumers don't immediately
         // claim the messages before they can be browsed.  This does not effect the buggy behavior.
/*
         consumer0 = session0.createConsumer(queue);
         consumer1 = session1.createConsumer(queue);
         consumer2 = session2.createConsumer(queue);
         consumer3 = session3.createConsumer(queue);
         consumer4 = session4.createConsumer(queue);
         consumer5 = session5.createConsumer(queue);
*/         
         // Step 7. We create anonymous message producer

         MessageProducer producer0 = session0.createProducer(null);

         // Step 8. We send 30 messages to the queue

         final int numMessages = 30;

         for (int i = 0; i < numMessages; i++)
         {
            TextMessage message0 = session0.createTextMessage("message " + i + " to session0");
            producer0.send(queue, message0);

            // See what's going where as messages are added
            browseSession(session0, queue, 0, i);
            browseSession(session1, queue, 1, i);
            browseSession(session2, queue, 2, i);
            browseSession(session3, queue, 3, i);
            browseSession(session4, queue, 4, i);
            browseSession(session5, queue, 5, i);
         }

         System.out.println("browse before creating consumers");
         browseSession(session0, queue, 0, -1);
         browseSession(session1, queue, 1, -1);
         browseSession(session2, queue, 2, -1);
         browseSession(session3, queue, 3, -1);
         browseSession(session4, queue, 4, -1);
         browseSession(session5, queue, 5, -1);

         // Step 6. <RELOCATED> We create queue consumers on all sessions

         consumer0 = session0.createConsumer(queue);
         consumer1 = session1.createConsumer(queue);
         consumer2 = session2.createConsumer(queue);
         consumer3 = session3.createConsumer(queue);
         consumer4 = session4.createConsumer(queue);
         consumer5 = session5.createConsumer(queue);

         // Notice that the consumers upon creation empty out their respective queue portions
         // before actually receiving anything!  (There is consumer caching that can be disabled, but at a performance penalty.)
         System.out.println("browse again after creating consumers");
         browseSession(session0, queue, 0, -2);
         browseSession(session1, queue, 1, -2);
         browseSession(session2, queue, 2, -2);
         browseSession(session3, queue, 3, -2);
         browseSession(session4, queue, 4, -2);
         browseSession(session5, queue, 5, -2);
         
         // Step 9. Verify all consumers receive their messages

         for (int i = 0; i < numMessages/6; i++)
         {
        	 if (pullMessage(consumer0, i, 0) == false) {return false;}
        	 if (pullMessage(consumer1, i, 1) == false) {return false;}
        	 if (pullMessage(consumer2, i, 2) == false) {return false;}
        	 if (pullMessage(consumer3, i, 3) == false) {return false;}
        	 if (pullMessage(consumer4, i, 4) == false) {return false;}
        	 if (pullMessage(consumer5, i, 5) == false) {return false;}
         }

         return true;
      }
      
      finally
      {
         // Step 10. Be sure to close our resources!

         if (connection0 != null)
         {
        	cleanUp(consumer0, " - consumer0 cleanup");
        	session0.close();  //grasping at straws
            connection0.close();
         }

         if (connection1 != null)
         {
         	cleanUp(consumer1, " - consumer1 cleanup");
         	session1.close();
            connection1.close();
         }

         if (connection2 != null)
         {
         	cleanUp(consumer2, " - consumer2 cleanup");
         	session2.close(); 
            connection2.close();
         }

         if (connection3 != null)
         {
         	cleanUp(consumer3, " - consumer3 cleanup");
         	session3.close(); 
            connection3.close();
         }

         if (connection4 != null)
         {
         	cleanUp(consumer4, " - consumer4 cleanup");
         	session4.close(); 
            connection4.close();
         }

         if (connection5 != null)
         {
         	cleanUp(consumer5, " - consumer5 cleanup");
         	session5.close(); 
            connection5.close();
         }

      }
      
   }
   public void browseSession(Session session, Queue queue, int sessionNum, int pass) throws Exception
   {
       // Create the JMS QueueBrowser
       QueueBrowser browser = session.createBrowser(queue);
       System.out.println("queue browser created - session" + sessionNum + " on pass:" + pass);

       // Browse the messages on the queue
       // Browsing a queue does not remove the messages from the queue
       Enumeration messageEnum = browser.getEnumeration(); 
//       System.out.println("browser hasMoreElements?" + messageEnum.hasMoreElements());
       while (messageEnum.hasMoreElements())
       {
          TextMessage message = (TextMessage)messageEnum.nextElement();
          System.out.println(message.getText() + " - browsed on session:" + sessionNum + " on pass:" + pass);
       }

       // Close the browser
       browser.close();      	  
   }
   
   public boolean pullMessage(MessageConsumer consumer, int msgNum, int consNum) throws Exception
   {
       TextMessage received = (TextMessage)consumer.receive(5000);

       if (received == null)
       {
          System.out.println("Failed attempting to receive message in loop " + msgNum + " from consumer:" + consNum);
          return false;
       }
       else
       {
       	  System.out.println(received.getText() + " from cons" + consNum);
       	  return true;
       }
	   
   }
   
   public void cleanUp(MessageConsumer consumer, String msg)
   {
	   try
	   {
		   TextMessage receiver = (TextMessage)consumer.receive(100);
		   Boolean queueNotEmpty = (receiver != null);
		   while(queueNotEmpty)
		   {
			   System.out.println(receiver.getText() + msg);
			   receiver = (TextMessage)consumer.receive(100);
			   queueNotEmpty = (receiver != null);
		   }
	   }
	   catch (JMSException e)
	   {
	   }
   }
}